
import os
from contextlib import redirect_stdout
import io
import time
import pandas as pd
import warnings
import traceback
import sys
import json
warnings.filterwarnings('ignore')

try:
    # Attempt package-relative import
    from . import models, prompts, func_calls, qa_retrieval, google_search, reg_ex, log_manager, output_manager, utils
except ImportError:
    # Fall back to script-style import
    import models, prompts, func_calls, qa_retrieval, google_search, reg_ex, log_manager, output_manager, utils

class BambooAI:
    def __init__(self, df: pd.DataFrame = None,
                 max_conversations: int = 4,
                 debug: bool = False, 
                 vector_db: bool = False, 
                 search_tool: bool = False,
                 exploratory: bool = True,
                 ):
        
        # Output
        self.output_manager = output_manager.OutputManager()

        # Check if the OPENAI_API_KEY environment variable is set
        if not os.getenv('OPENAI_API_KEY'):
            raise EnvironmentError("OPENAI_API_KEY environment variable not found.")
        
        # Check if the SERPER_API_KEY environment variable is set
        if not os.getenv('SERPER_API_KEY'):
            self.output_manager.print_wrapper("Warning: SERPER_API_KEY environment variable not found. Disabling google_search.")
            search_tool = False
        
        # Check if the PINECONE_API_KEY and PINECONE_ENV environment variables are set if vector_db is True
        if vector_db:
            PINECONE_API_KEY = os.getenv('PINECONE_API_KEY')
            PINECONE_ENV = os.getenv('PINECONE_ENV')
            
            if PINECONE_API_KEY is None or PINECONE_ENV is None:
                self.output_manager.print_wrapper("Warning: PINECONE_API_KEY or PINECONE_ENV environment variable not found. Disabling vector_db.")
                vector_db = False

        self.MAX_ERROR_CORRECTIONS = 5
        # Set the maximum number of question/answer pairs to be kept in the conversation memmory
        self.MAX_CONVERSATIONS = (max_conversations*2) - 1
        
        # Dataframe
        self.df = df if df is not None else None
        
        # Results of the code execution
        self.code_exec_results = None

        # Set the debug mode. This mode is True when you want the model to debug the code and correct it.
        self.debug = debug
        # Set the vector_db mode. This mode is True when you want the model to rank the generated code, and store the results above threshold in a vector database.
        self.vector_db = vector_db

        # Set the exploratory mode. This mode is True when you want the model to evaluate the original user prompt and break it down in algorithm.
        self.exploratory = exploratory
        
        # Prompts
        # Define list of templates
        templates = [
            "default_example_output_df",
            "default_example_output_gen",
            "default_example_plan_df",
            "default_example_plan_gen",
            "expert_selector_system",
            "expert_selector_user",
            "analyst_selector_system",
            "analyst_selector_user",
            "planner_system",
            "planner_user_gen",
            "planner_user_df",
            "theorist_system",
            "google_search_query_generator_system",
            "google_search_react_system",
            "code_generator_system_df",
            "code_generator_system_gen",
            "code_generator_user_df",
            "code_generator_user_gen",
            "error_corector_system",
            "code_debugger_system",
            "code_ranker_system",
            "solution_summarizer_system"
        ]

        prompt_data = {}

        # Check if the JSON file exists
        if os.path.exists("PROMPT_TEMPLATES.json"):
            # Load from JSON file
            with open("PROMPT_TEMPLATES.json", "r") as f:
                prompt_data = json.load(f)

        # Set templates to the values from the JSON file or the default values. This dynamicaly sets the object attributes.
        # These attributes are part of the object's state and will exist as long as the object itself exists.
        # The attributes can be called using self.<attribute_name> throughout the class.
        for template in templates:
            value = prompt_data.get(template, getattr(prompts, template, ""))
            setattr(self, template, value)

        # Regular expresions
        self._extract_code = reg_ex._extract_code
        self._extract_rank = reg_ex._extract_rank
        self._extract_expert = reg_ex._extract_expert
        self._extract_analyst = reg_ex._extract_analyst
        self._extract_plan = reg_ex._extract_plan
        self._remove_examples = reg_ex._remove_examples

        # Functions
        self.task_eval_function = func_calls.task_eval_function
        self.insights_function = func_calls.solution_insights_function
        self.search_function = func_calls.search_function

        # LLM calls
        self.llm_call = models.llm_call
        self.llm_stream = models.llm_stream

        # Logging
        self.token_cost_dict = {
                                'gpt-3.5-turbo-0613': {'prompt_tokens': 0.0015, 'completion_tokens': 0.0020},
                                'gpt-3.5-turbo-1106': {'prompt_tokens': 0.0010, 'completion_tokens': 0.0020},
                                'gpt-3.5-turbo-instruct': {'prompt_tokens': 0.0015, 'completion_tokens': 0.0020},
                                'gpt-3.5-turbo-16k': {'prompt_tokens': 0.0030, 'completion_tokens': 0.0040},
                                'gpt-4': {'prompt_tokens': 0.03, 'completion_tokens': 0.06},
                                'gpt-4-1106-preview': {'prompt_tokens': 0.01, 'completion_tokens': 0.03},
                                'gpt-4-0613': {'prompt_tokens': 0.03, 'completion_tokens': 0.06}, 
                                'gpt-4-0125-preview': {'prompt_tokens': 0.01, 'completion_tokens': 0.03},
                                'gpt-4-turbo': {'prompt_tokens': 0.01, 'completion_tokens': 0.03},
                                'llama3-70b-8192': {'prompt_tokens': 0.00059, 'completion_tokens': 0.00079}, #Groq Llama3
                                'gemini-1.5-pro-latest': {'prompt_tokens': 0.0025, 'completion_tokens': 0.0025}, 
                                'claude-3-haiku-20240307': {'prompt_tokens': 0.00025, 'completion_tokens': 0.00079}, 
                                'claude-3-sonnet-20240229': {'prompt_tokens': 0.003, 'completion_tokens': 0.015},
                                'claude-3-opus-20240307': {'prompt_tokens': 0.015, 'completion_tokens': 0.075},
                                'mistral-small': {'prompt_tokens': 0.002, 'completion_tokens': 0.006},
                                'open-mixtral-8x22b': {'prompt_tokens': 0.002, 'completion_tokens': 0.006},
                                'mistral-large': {'prompt_tokens': 0.008, 'completion_tokens': 0.024},
                                }
        self.log_and_call_manager = log_manager.LogAndCallManager(self.token_cost_dict)
        self.chain_id = None

        # Messages lists
        self.pre_eval_messages = [{"role": "system", "content": self.expert_selector_system}]
        self.select_analyst_messages = [{"role": "system", "content": self.analyst_selector_system}]
        self.eval_messages = [{"role": "system", "content": self.planner_system.format(utils.get_readable_date())}]
        self.code_messages = [{"role": "system", "content": self.code_generator_system_df}]

        # QA Retrieval
        self.add_question_answer_pair = qa_retrieval.add_question_answer_pair
        self.retrieve_answer = qa_retrieval.retrieve_answer
        self.similarity_threshold = 0.8

        # Google Search
        self.search_tool = search_tool
        self.google_search = google_search.SmartSearchOrchestrator()

    ######################
    ### Util Functions ###
    ######################

    def reset_messages_and_logs(self):
        self.pre_eval_messages = [{"role": "system", "content": self.expert_selector_system}]
        self.select_analyst_messages = [{"role": "system", "content": self.analyst_selector_system}]
        self.eval_messages = [{"role": "system", "content": self.planner_system.format(utils.get_readable_date())}]
        self.code_messages = [{"role": "system", "content": self.code_generator_system_df}]
        self.code_exec_results = None

        self.log_and_call_manager.clear_run_logs()
    
    ######################
    ### Eval Functions ###
    ######################
    
    def select_expert(self, pre_eval_messages):
        '''Call the Expert Selector'''
        agent = 'Expert Selector'
        using_model,provider = models.get_model_name(agent)

        self.output_manager.display_tool_start(agent,using_model)

        # Call OpenAI API to evaluate the task
        llm_response = self.llm_stream(self.log_and_call_manager,pre_eval_messages, agent=agent,chain_id=self.chain_id)
        expert, requires_dataset, confidence = self._extract_expert(llm_response)

        return expert, requires_dataset, confidence
    
    def select_analyst(self, select_analyst_messages):
        '''Call the Analyst Selector'''
        agent = 'Analyst Selector'
        # Call OpenAI API to evaluate the task
        llm_response = self.llm_stream(self.log_and_call_manager,select_analyst_messages, agent=agent, chain_id=self.chain_id)
        analyst, rephrased_query = self._extract_analyst(llm_response)

        return analyst, rephrased_query
    
    def task_eval(self, eval_messages, agent):
        '''Call the Task Evaluator'''
        using_model,provider = models.get_model_name(agent)

        self.output_manager.display_tool_start(agent,using_model)

        if self.search_tool:
            tools=self.search_function
        else:
            tools=None

        # Call OpenAI API to evaluate the task
        llm_response = self.llm_stream(self.log_and_call_manager,eval_messages, agent=agent, chain_id=self.chain_id, tools=tools)
        if agent == 'Planner':
            response = self._extract_plan(llm_response)
        else:
            response = llm_response
            
        return response
    
    def taskmaster(self, question, df_columns):
        '''Taskmaster function to select the expert, refine the expert selection, and formulate a task for the expert'''
        task = None
        analyst = None
        rephrased_query = None
        requires_dataset = None
        confidence = None

        ######## Select Expert ###########
        self.pre_eval_messages.append({"role": "user", "content": self.expert_selector_user.format(question)})
        expert,requires_dataset,confidence  = self.select_expert(self.pre_eval_messages) 
        self.pre_eval_messages.append({"role": "assistant", "content": f"expert:{expert},requires_dataset:{requires_dataset},confidence:{confidence}"})

        ######## Refine Expert Selection, and Formulate the task for the expert ###########
        if expert == 'Data Analyst':
            self.select_analyst_messages.append({"role": "user", "content": self.analyst_selector_user.format(question, None if self.df is None else self.df.columns.tolist())})
            analyst, rephrased_query = self.select_analyst(self.select_analyst_messages)
            self.select_analyst_messages.append({"role": "assistant", f"content": f"analyst:{analyst},rephrased_query:{rephrased_query}"})

            # Retrieve the matching code and plan from the vector database if exists
            if self.vector_db:
                example_code, example_plan = self.retrieve_answer(rephrased_query, df_columns, similarity_threshold=self.similarity_threshold)

                if example_plan is not None:
                    example_plan = f"Use the below plan as base for your answer:\n\n```yaml\n{example_plan}\n```"
                else:     
                    if analyst == 'Data Analyst DF':
                        example_plan = self.default_example_plan_df
                    elif analyst == 'Data Analyst Generic':
                        example_plan = self.default_example_plan_gen

            if analyst == 'Data Analyst DF':
                self.eval_messages.append({"role": "user", "content": self.planner_user_df.format(None if self.df is None else self.df.dtypes.to_string(max_rows=None), example_plan, question)}) 
                # Replace first dict in messages with a new system task. This is to distinguish between the two types of analysts
                self.code_messages[0] = {"role": "system", "content": self.code_generator_system_df}
            elif analyst == 'Data Analyst Generic':
                self.eval_messages.append({"role": "user", "content": self.planner_user_gen.format(example_plan, question)})
                # Replace first dict in messages with a new system task. This is to distinguish between the two types of analysts
                self.code_messages[0] = {"role": "system", "content": self.code_generator_system_gen}
            agent = 'Planner'

        elif expert == 'Data Analysis Theorist':
            self.eval_messages.append({"role": "user", "content": self.theorist_system.format(question)})
            agent = 'Theorist'
        else:
            self.eval_messages.append({"role": "user", "content": self.theorist_system.format(question)})

        task_eval = self.task_eval(self.eval_messages, agent)
        self.eval_messages.append({"role": "assistant", "content": task_eval})

        # Remove the oldest conversation from the eval_messages list
        if len(self.eval_messages) > self.MAX_CONVERSATIONS:
            self.eval_messages.pop(1)
            self.eval_messages.pop(1)

        if expert == 'Data Analysis Theorist':
            self.log_and_call_manager.print_summary_to_terminal()
        elif expert == 'Data Analyst':
            task = task_eval
        else:
            self.log_and_call_manager.print_summary_to_terminal()

        return analyst, task, rephrased_query, requires_dataset, confidence
    
    #####################
    ### Main Function ###
    #####################

    def pd_agent_converse(self, question=None):
      
        # Initialize the loop variable. If user provided question as an argument, the loop will finish after one iteration
        if question is not None:
            loop = False
        else:
            loop = True

        # Set the chain id
        chain_id = int(time.time())
        self.chain_id = chain_id

        # Reset messages and logs
        self.reset_messages_and_logs()

        # Start the conversation loop
        while True:
            if loop:
                question = self.output_manager.display_user_input_prompt()
                # If the user types 'exit', break out of the loop
                if question.strip().lower() == 'exit':
                    self.log_and_call_manager.consolidate_logs()    
                    break
                
            if self.exploratory is True:
                # Call the taskmaister method with the user's question if the exploratory mode is True
                analyst, task, rephrased_query, requires_dataset, confidence = self.taskmaster(question,'' if self.df is None else self.df.columns.tolist())
                if not loop:
                    if not analyst:
                        self.log_and_call_manager.consolidate_logs()
                        return
                else:
                    if not analyst:
                        continue
            else:
                analyst = 'Data Analyst DF'
                task = question
            
            if self.vector_db:
                # Call the retrieve_answer method to check if the question has already been asked and answered
                if analyst == 'Data Analyst DF':
                    df_columns = '' if self.df is None else self.df.columns.tolist()
                elif analyst == 'Data Analyst Generic':
                    df_columns = ''

                example_code, example_plan = self.retrieve_answer(rephrased_query, df_columns, similarity_threshold=self.similarity_threshold)
                if example_code is not None:
                    example_code = f"Review the the below code, and use as base for your answer:\n\n```python\n{example_code}\n```"
                else:     
                    if analyst == 'Data Analyst DF':
                        example_code = self.default_example_output_df
                    else:
                        example_code = self.default_example_output_gen
            else:
                if analyst == 'Data Analyst DF':
                    example_code = self.default_example_output_df
                else:
                    example_code = self.default_example_output_gen

            # Call the generate_code() method to genarate and debug the code
            code = self.generate_code(analyst, task, self.code_messages, example_code)
            # Call the execute_code() method to execute the code and summarise the results
            answer, results, code = self.execute_code(analyst,code, task, question, self.code_messages)

            # Rank the LLM response
            if self.vector_db:
                rank = self.rank_code(results, code,task)
            else:
                rank = ""

            # Display the results
            self.output_manager.display_results(df=self.df, answer=answer, code=code, rank=rank, vector_db=self.vector_db)

            if self.vector_db:
                rank_feedback = self.output_manager.display_user_input_rank()

                # If the user types "yes", use the rank as is. If not, use the user's rank.
                if rank_feedback.strip().lower() == 'yes':
                    rank = rank
                elif rank_feedback in map(str, range(0, 11)):
                    rank = rank_feedback
                else:
                    rank = rank

                # Add the question and answer pair to the QA retrieval index
                self.add_question_answer_pair(rephrased_query, task, df_columns, code, rank)

            self.log_and_call_manager.print_summary_to_terminal()
            
            if not loop:
                self.log_and_call_manager.consolidate_logs()
                return 
            
    ######################
    ### Code Functions ###
    ######################
            
    def generate_code(self, analyst, task, code_messages, example_code):
        agent = 'Code Generator'
        # Add a user message with the updated task prompt to the messages list
        if analyst == 'Data Analyst DF':
            code_messages.append({"role": "user", "content": self.code_generator_user_df.format(None if self.df is None else self.df.dtypes.to_string(max_rows=None), task, self.code_exec_results, example_code)})
        elif analyst == 'Data Analyst Generic':
            code_messages.append({"role": "user", "content": self.code_generator_user_gen.format(task, self.code_exec_results, example_code)})

        using_model,provider = models.get_model_name(agent)

        self.output_manager.display_tool_start(agent,using_model)

        # Call the OpenAI API or a local code model
        llm_response = self.llm_stream(self.log_and_call_manager,code_messages, agent=agent, chain_id=self.chain_id)
        code_messages.append({"role": "assistant", "content": llm_response})

        # Extract the code from the API response
        code = self._extract_code(llm_response,analyst,provider)

        # Debug code if debug parameter is set to True
        if self.debug:
            code = self.debug_code(analyst, code, task)
        else:
            self.output_manager.display_tool_end(agent)

        return code
    
    def debug_code(self,analyst,code,question):
        agent = 'Code Debugger'
        # Initialize the messages list with a system message containing the task prompt
        debug_messages = [{"role": "user", "content": self.code_debugger_system.format(code,question)}]
        
        using_model,provider = models.get_model_name(agent)

        self.output_manager.display_tool_start(agent,using_model)

        # Call the OpenAI API
        llm_response = self.llm_stream(self.log_and_call_manager, debug_messages, agent=agent, chain_id=self.chain_id)
        
        # Extract the code from the API response
        debugged_code = self._extract_code(llm_response,analyst,provider)       
        self.output_manager.display_tool_end(agent)

        return debugged_code

    def execute_code(self, analyst, code, task, original_question, code_messages):
        agent = 'Code Executor'
        # Initialize error correction counter
        error_corrections = 0

        # Create a copy of the original self.df
        if self.df is not None:
            original_df = self.df.copy()

        # Redirect standard output to a StringIO buffer
        with redirect_stdout(io.StringIO()) as output:
            # Try to execute the code and handle errors
            while error_corrections < self.MAX_ERROR_CORRECTIONS:
                try:
                    # Remove the oldest conversation from the messages list
                    if len(code_messages) > self.MAX_CONVERSATIONS:
                        code_messages.pop(1)
                        code_messages.pop(1)

                    # Execute the code
                    if code is not None:
                        exec(code, {'df': self.df})
                        # Remove examples from the messages list to minimize the number of tokens used
                        self.code_messages = self._remove_examples(self.code_messages)
                    break
                except Exception as error:
                    # Capture the full traceback
                    exc_type, exc_value, tb = sys.exc_info()
                    full_traceback = traceback.format_exc()
                    # Filter the traceback
                    exec_traceback = self.filter_exec_traceback(full_traceback, exc_type.__name__, str(exc_value)) 

                    # Increment the error corrections counter
                    error_corrections += 1

                    # Reset df to the original state before trying again
                    if self.df is not None:
                        self.df = original_df.copy()

                    code, code_messages = self.correct_code_errors(exec_traceback, error_corrections, code_messages, analyst)
              
        # Get the output from the executed code
        results = output.getvalue()
        
        # Store the results in a class variable so it can be appended to the subsequent messages list
        self.code_exec_results = results

        summary = self.summarise_solution(task, original_question, results)

        # Reset the StringIO buffer
        output.truncate(0)
        output.seek(0)

        return summary, results, code
    
    def filter_exec_traceback(self, full_traceback, exception_type, exception_value):
        # Split the full traceback into lines and filter those that originate from "<string>"
        filtered_tb_lines = [line for line in full_traceback.split('\n') if '<string>' in line]

        # Combine the filtered lines and append the exception type and message
        filtered_traceback = '\n'.join(filtered_tb_lines)
        if filtered_traceback:  # Add a newline only if there's a traceback to show
            filtered_traceback += '\n'
        filtered_traceback += f"{exception_type}: {exception_value}"

        return filtered_traceback
    
    def correct_code_errors(self, error, error_corrections, code_messages, analyst):
        agent = 'Error Corrector'

        model,provider = models.get_model_name(agent)

        #If error correction is greater than 2 remove the first error correction
        if error_corrections > 2:
            del code_messages[-4] 
            del code_messages[-3]
        
        # Append the error message to the messages list
        code_messages.append({"role": "user", "content": self.error_corector_system.format(error)})

        # Display the error message
        self.output_manager.display_error(error)

        llm_response = self.llm_call(self.log_and_call_manager,code_messages,agent=agent, chain_id=self.chain_id)
        code_messages.append({"role": "assistant", "content": llm_response})
        code = self._extract_code(llm_response,analyst,provider)

        return code, code_messages

    def rank_code(self,results, code,question):
        agent = 'Code Ranker'
        # Initialize the messages list with a user message containing the task prompt
        rank_messages = [{"role": "user", "content": self.code_ranker_system.format(code,results,question)}]

        using_model,provider = models.get_model_name(agent)

        self.output_manager.display_tool_start(agent,using_model)

        # Call the OpenAI API 
        llm_response = self.llm_call(self.log_and_call_manager,rank_messages,agent=agent, chain_id=self.chain_id)

        # Extract the rank from the API response
        rank = self._extract_rank(llm_response)       

        return rank
    
    ############################
    ## Summarise the solution ##
    ############################

    def summarise_solution(self, task, original_question, results):
        agent = 'Solution Summarizer'

        # Initialize the messages list with a user message containing the task prompt
        insights_messages = [{"role": "user", "content": self.solution_summarizer_system.format(original_question, task, results)}]
        # Call the OpenAI API
        summary = self.llm_call(self.log_and_call_manager,insights_messages,agent=agent, chain_id=self.chain_id)

        return summary


import importlib
import os
import time
import json


def load_llm_config():

    default_llm_config = [
    {"agent": "Expert Selector", "details": {"model": "gpt-4-turbo", "provider":"openai","max_tokens": 500, "temperature": 0}},
    {"agent": "Analyst Selector", "details": {"model": "gpt-4-turbo", "provider":"openai","max_tokens": 500, "temperature": 0}},
    {"agent": "Theorist", "details": {"model": "gpt-4-turbo", "provider":"openai","max_tokens": 2000, "temperature": 0}},
    {"agent": "Planner", "details": {"model": "gpt-4-turbo", "provider":"openai","max_tokens": 2000, "temperature": 0}},
    {"agent": "Code Generator", "details": {"model": "gpt-4-turbo", "provider":"openai","max_tokens": 2000, "temperature": 0}},
    {"agent": "Code Debugger", "details": {"model": "gpt-4-turbo", "provider":"openai","max_tokens": 2000, "temperature": 0}},
    {"agent": "Error Corrector", "details": {"model": "gpt-4-turbo", "provider":"openai","max_tokens": 2000, "temperature": 0}},
    {"agent": "Code Ranker", "details": {"model": "gpt-4-turbo", "provider":"openai","max_tokens": 500, "temperature": 0}},
    {"agent": "Solution Summarizer", "details": {"model": "gpt-4-turbo", "provider":"openai","max_tokens": 2000, "temperature": 0}},
    {"agent": "Google Search Query Generator", "details": {"model": "gpt-4-turbo", "provider":"openai","max_tokens": 2000, "temperature": 0}},
    {"agent": "Google Search Summarizer", "details": {"model": "gpt-4-turbo", "provider":"openai","max_tokens": 2000, "temperature": 0}}
    ]

    # Get the LLM_CONFIG environment variable
    if os.environ.get('LLM_CONFIG'):
        llm_config = os.environ.get('LLM_CONFIG')
        llm_config = json.loads(llm_config)
    # load config from the JSON file
    elif os.path.exists("LLM_CONFIG.json"):
        try:
            with open("LLM_CONFIG.json", 'r') as f:
                llm_config = json.load(f)
        except Exception as e:
            llm_config = default_llm_config
    # Use hardcoded default configuration
    else:
        llm_config = default_llm_config

    return llm_config

def init(agent):
    
    llm_config = load_llm_config()

    for item in llm_config:
        if item['agent'] == agent:
            details = item.get('details', {})
            model = details.get('model', 'Unknown')
            provider = details.get('provider', 'Unknown')
            max_tokens = details.get('max_tokens', 'Unknown')
            temperature = details.get('temperature', 'Unknown')

    return model, provider, max_tokens, temperature

def get_model_name(agent):
    
    llm_config = load_llm_config()

    for item in llm_config:
        if item['agent'] == agent:
            details = item.get('details', {})
            model = details.get('model', 'Unknown')
            provider = details.get('provider', 'Unknown')

    return model, provider

def try_import(module_name):
    try:
        # Attempt package-relative import
        module = importlib.import_module(f'.{module_name}','bambooai')
    except ImportError:
        # Fall back to script-style import
        module = importlib.import_module(module_name)
    return module

def llm_call(log_and_call_manager, messages: str, agent: str = None, chain_id: str = None):
    timestamp = time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())
    model, provider, max_tokens, temperature = init(agent)

    # Map providers to their respective function names ('llm_stream' for local, 'llm_call' for others)
    provider_function_map = {
        'local': 'llm_stream',
        'groq': 'llm_call',
        'openai': 'llm_call',
        'ollama': 'llm_call',
        'gemini': 'llm_call',
        'anthropic': 'llm_call',
        'mistral': 'llm_call'
    }

    if provider in provider_function_map:
        # Try to import the correct module
        provider_module = try_import(f'{provider}_models')

        # Call the appropriate function from the imported module
        function_name = provider_function_map[provider]
        content_received, local_llm_messages, prompt_tokens_used, completion_tokens_used, total_tokens_used, elapsed_time, tokens_per_second = getattr(provider_module, function_name)(messages, model, temperature, max_tokens)
        
        # Log the results
        log_and_call_manager.write_to_log(agent, chain_id, timestamp, model, local_llm_messages, content_received, prompt_tokens_used, completion_tokens_used, total_tokens_used, elapsed_time, tokens_per_second)
        
        return content_received
    else:
        raise ValueError(f"Unsupported provider: {provider}")

def llm_stream(log_and_call_manager, messages: str, agent: str = None, chain_id: str = None, tools:str = None):
    timestamp = time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())

    # Initialize the LLM parameters
    model, provider, max_tokens, temperature = init(agent)

    # Map providers to their respective function names
    provider_function_map = {
        'local': 'llm_stream',
        'groq': 'llm_stream',
        'openai': 'llm_stream',
        'ollama': 'llm_stream',
        'gemini': 'llm_stream',
        'anthropic': 'llm_stream',
        'mistral': 'llm_stream'
    }

    if provider in provider_function_map:
        # Try to import the correct module
        provider_module = try_import(f'{provider}_models')

        # Call the appropriate function from the imported module
        function_name = provider_function_map[provider]
        content_received, local_llm_messages, prompt_tokens_used, completion_tokens_used, total_tokens_used, elapsed_time, tokens_per_second = getattr(provider_module, function_name)(log_and_call_manager, chain_id, messages, model, temperature, max_tokens, tools)
        
        # Log the results
        log_and_call_manager.write_to_log(agent, chain_id, timestamp, model, local_llm_messages, content_received, prompt_tokens_used, completion_tokens_used, total_tokens_used, elapsed_time, tokens_per_second)
        
        return content_received
    else:
        raise ValueError(f"Unsupported provider: {provider}")

        


# func_calls.py

task_eval_function = [
    {
        "name": "QA_Response",
        "description": "The answer classification function",
        "parameters": {
            "type": "object",
            "properties": {
                "answer": {
                    "type": "string",
                    "description": "Answer as a narrative, or as an algorithm",
                },
                "answer_type": {
                    "type": "string",
                    "enum": ["narrative", "algorithm"],
                    "description": "Task classification",
                },
            },
            "required": ["answer", "answer_type"],
            },
    },
    ]

solution_insights_function = [
    {
        "name": "Solution_Insights",
        "description": "The solution summary and insights function",
        "parameters": {
            "type": "object",
            "properties": {
                "insight": {
                    "type": "string",
                    "description": "Delivers a comprehensive summary of the outcomes obtained from the execution of the python code",
                },
            },
            "required": ["insight"],
            },
    },
    ]

search_function = [
  {
    "type": "function",
    "function": {
      "name": "google_search",
      "description": "This function executes a Google search and delivers results by identifying and summarizing the most relevant documents from the returned search results.",
      "parameters": {
        "type": "object",
        "properties": {
          "search_query": {
            "type": "string",
            "description": "A query string to search for",
          }
        },
        "required": ["search_result"]
      }
    }
  }
]


import openai
import re
from serpapi import GoogleSearch
import json
import numpy as np
import requests
import os
from newspaper import Article

openai_client = openai.OpenAI()

MAX_ITERATIONS = 5
CHUNK_SIZE = 512
TOP_K_RESULTS = 6
SEARCH_RESULTS = 20
NUM_DOCUMENTS = 30

class ChatBot:
    def __init__(self):
        self.agent = 'Google Search Query Generator'
        self.completion = None
        
    def __call__(self, log_and_call_manager, chain_id, messages):
        result = self.execute(log_and_call_manager, chain_id, messages)
        return result

    def execute(self,log_and_call_manager, chain_id, messages):
        try:
            # Attempt package-relative import
            from . import models
        except ImportError:
            # Fall back to script-style import
            import models

        self.completion = models.llm_stream(log_and_call_manager,messages, agent=self.agent, chain_id=chain_id)

        return self.completion

class SmartSearchOrchestrator:
    action_re = re.compile(r'^Action: (\w+): (.*)$')

    def __init__(self, log_and_call_manager=None,chain_id=None, messages=None):
        self.log_and_call_manager = log_and_call_manager
        self.chain_id = chain_id
        self.messages = messages
     
        self.chat_bot = ChatBot() 
        self.google_search = GoogleSearch()
        self.calculate = Calculator()

        self.known_actions = {
            "google_search": self.google_search,
            "calculate": self.calculate,
}

    def perform_query(self, log_and_call_manager, chain_id, messages, max_turns=MAX_ITERATIONS):
        try:
            # Attempt package-relative import
            from . import output_manager
        except ImportError:
            # Fall back to script-style import
            import output_manager

        output_handler = output_manager.OutputManager()

        i = 0
        observation = None
        next_prompt = messages
        while i < max_turns:
            i += 1
            result = self.chat_bot(log_and_call_manager, chain_id, next_prompt)
            next_prompt.append({"role": "assistant", "content": result})
            actions = [self.action_re.match(a) for a in result.split('\n') if self.action_re.match(a)]
            if actions:
                # There is an action to run
                action, action_input = actions[0].groups()
                if action not in self.known_actions:
                    raise Exception("Unknown action: {}: {}".format(action, action_input))
                output_handler.display_search_task(action, action_input)
                observation, links = self.known_actions[action](log_and_call_manager, chain_id, action_input)
                if links:
                    for link in links:
                        output_handler.print_wrapper(f"\nTitle: {link['title']}\nLink: {link['link']}")
                output_handler.print_wrapper("\nObservation:", observation)
                next_prompt.append({"role": "user", "content": "Observation: {}".format(observation)})
            else:
                break
        
        if observation is None:
            observation = result

        return observation
    
    def __call__(self, log_and_call_manager, chain_id, messages):
        return self.perform_query(log_and_call_manager, chain_id, messages)
    
### SEARCH ACTIONS ###

# Define a class to perform a Google search and retrieve the content of the resulting pages    
class SearchEngine:
    # Perform a Google search using the SERPer API
    def search_google(self, query, gl='us', hl='en'):
        url = "https://google.serper.dev/search"
        payload = json.dumps({"q": query, "gl": gl, "hl": hl, "num": SEARCH_RESULTS, "autocorrect": False})
        headers = {'X-API-KEY': os.environ['SERPER_API_KEY'], 'Content-Type': 'application/json'}

        response = requests.request("POST", url, headers=headers, data=payload)
        response = json.loads(response.text)
        return response
    
    # Download and parse an article from a URL using the Newspaper library
    def search_url(self, url, document_size=CHUNK_SIZE):
        try:
            article = Article(url)
            article.download()
            article.parse()
        except:
            return []

        full_text = article.text.replace('\n', ' ')
        full_words = full_text.split(' ')
        # Create a list of "documents". Each "document" is a string that contains document_size consecutive words from the article.
        documents = [' '.join(full_words[i:i+document_size]) for i in range(0, len(full_words), document_size)]
        # Remove documents that are too short
        documents = [doc for doc in documents if len(doc) > 100]
        return documents
    
    # Perform a Google search and retrieve the content of the top results. Maximum word count is num_documents * context_size (default 7680)
    def __call__(self, query, num_documents=NUM_DOCUMENTS):
        google_resp = self.search_google(query)
        
        documents = []
        top_links = []

        for i, resp in enumerate(google_resp['organic']):
            documents += self.search_url(resp['link'])
            if i < 5:
                top_links.append({'title': resp['title'], 'link': resp['link']})
            if len(documents) > num_documents:
                break

        documents = documents[:num_documents]
        return documents, top_links

# Define a class to retrieve the most relevant documents for a question
class DocumentRetriever:
    # Create a vector embedding of a text using OpenAI's 'text-embedding-3-large' model
    def encode(self, input):
        resp = openai_client.embeddings.create(
            model = 'text-embedding-3-large',
            input = input
        )

        embeds = np.array([d.embedding for d in resp.data])
        return embeds
    
    # Retrieve the most relevant documents for a question using vector embeddings
    def __call__(self, question, documents, k=TOP_K_RESULTS):
        question_embed = self.encode(question)
        document_embeds = self.encode(documents)

        # Compute the cosine similarity between the question embedding and the document embeddings
        scores = np.dot(question_embed, document_embeds.T)
        # Sort the documents by their cosine similarity to the question
        ranks = np.argsort(-scores)[0, :k].tolist()
        # Retrieve the most relevant documents
        contexts = [documents[r] for r in ranks]
        return contexts

# Define a class to generate an answer to a question based on a set of documents
class Reader:
    def __call__(self,log_and_call_manager,chain_id,query, contexts):
        agent = 'Google Search Summarizer'
        text = ""
        
        try:
            # Attempt package-relative import
            from . import models, prompts
        except ImportError:
            # Fall back to script-style import
            import models, prompts
        
        # Construct prompt and messages
        for ctx in contexts:
            text += f'* {ctx}\n'

        # Check if PROMPT_TEMPLATES.json exists and load the prompts from there. If not, use the default prompts.
        if os.path.exists("PROMPT_TEMPLATES.json"):
            # Load from JSON file
            with open("PROMPT_TEMPLATES.json", "r") as f:
                prompt_data = json.load(f)
            prompt = prompt_data.get("google_search_summarizer_system", "")
            prompt = prompt.format(query,text)
        else:
            prompt = prompts.google_search_summarizer_system.format(query,text)
            
        search_messages = [{"role": "system", "content": prompt}]

        llm_response = models.llm_stream(log_and_call_manager,search_messages, agent=agent, chain_id=chain_id)

        return llm_response
    
class GoogleSearch:
    def __init__(self):
        self.search_engine = SearchEngine()
        self.document_retriever = DocumentRetriever()
        self.reader = Reader()

    def _extract_search_query(self,question: str) -> str:
        search_query = re.sub('\'|"', '',  question).strip()
        return search_query

    def __call__(self, log_and_call_manager,chain_id,question):
        question=self._extract_search_query(question)
        documents,top_links = self.search_engine(question)
        contexts = self.document_retriever(question, documents)
        answer = self.reader(log_and_call_manager,chain_id,question, contexts)

        return answer, top_links
    
### END SEARCH ACTIONS ###
    
class Calculator:
    def __call__(self, log_and_call_manager, chain_id, code):
        links = None
        # It's still important to be careful with eval(), as it can execute arbitrary code.
        try:
            return str(eval(code)),links
        except Exception as e:
            return str(e),links  # Return the error message as a string



import time
from ollama import Client
import tiktoken

client = Client()

try:
    # Attempt package-relative import
    from . import output_manager
except ImportError:
    # Fall back to script-style import
    import output_manager

output_manager = output_manager.OutputManager()

def llm_call(messages: str,model: str,temperature: str,max_tokens: str):  
    client = Client(host='http://localhost:11434')

    start_time = time.time()

    response = client.chat(
        model=model, 
        messages=messages,
        options = {
            'temperature': temperature,
            'top_k': 10,
        }, 
    )

    end_time = time.time()

    content = response['message']['content']

    elapsed_time = end_time - start_time

    prompt_tokens_used = response['prompt_eval_count']
    completion_tokens_used = response['eval_count']
    total_tokens_used = prompt_tokens_used + completion_tokens_used
    
    if elapsed_time > 0:
        tokens_per_second = completion_tokens_used / elapsed_time
    else:
        tokens_per_second = 0

    return content, messages, prompt_tokens_used, completion_tokens_used, total_tokens_used, elapsed_time, tokens_per_second

def llm_stream(log_and_call_manager, chain_id: str, messages: str,model: str,temperature: str,max_tokens: str,tools: str = None):  
    collected_chunks = []
    collected_messages = []

    client = Client(host='http://localhost:11434')
    response = client.chat(
        model=model, 
        messages=messages,
        stream =True,
        options = {
            'temperature': temperature,
            'top_k': 10,
        }, 
    )

    start_time = time.time()
    for chunk in response:
        collected_chunks.append(chunk)
        chunk_message = chunk['message']['content']
        collected_messages.append(chunk_message)
        output_manager.print_wrapper(chunk_message, end='', flush=True)  

    end_time = time.time()
    elapsed_time = end_time - start_time

    output_manager.print_wrapper("")

    full_reply_content = ''.join([m for m in collected_messages])

    # count the number of response tokens used
    completion_tokens_used = len(collected_chunks)

    # count the number of prompt tokens used
    encoding = tiktoken.encoding_for_model("gpt-4")   
    tokens_per_message = 3
    tokens_per_name = 1
    prompt_tokens_used = 0
    for message in messages:
        prompt_tokens_used += tokens_per_message
        for key, value in message.items():
            prompt_tokens_used += len(encoding.encode(value))
            if key == "name":
                prompt_tokens_used += tokens_per_name
    prompt_tokens_used += 3  # every reply is primed with <|start|>assistant<|message|>

    # calculate the total tokens used
    total_tokens_used = prompt_tokens_used + completion_tokens_used
    
    if elapsed_time > 0:
        tokens_per_second = completion_tokens_used / elapsed_time
    else:
        tokens_per_second = 0

    return full_reply_content, messages, prompt_tokens_used, completion_tokens_used, total_tokens_used, elapsed_time, tokens_per_second



import json
import pinecone
from sentence_transformers import SentenceTransformer
import os
import hashlib

try:
    # Attempt package-relative import
    from . import output_manager
except ImportError:
    # Fall back to script-style import
    import output_manager

output_manager = output_manager.OutputManager()

def init_pinecone():
    # Get the PINECONE_API_KEY and PINECONE_ENV environment variables
    PINECONE_API_KEY = os.getenv('PINECONE_API_KEY')
    PINECONE_ENV = os.getenv('PINECONE_ENV')

    if PINECONE_API_KEY is None or PINECONE_ENV is None:
        output_manager.print_wrapper("Warning: PINECONE_API_KEY or PINECONE_ENV environment variable not found.")
        return None, None

    # Initialize Pinecone
    pinecone.init(api_key=PINECONE_API_KEY,environment=PINECONE_ENV)

    # Load sentence transformer model
    model = SentenceTransformer('all-MiniLM-L6-v2')

    # Create a new Pinecone index if it doesnt exist
    index_name = "bambooai-qa-retrieval"
    if index_name not in pinecone.list_indexes():
        output_manager.print_wrapper(f"Creating a new vector db index. Please wait... {index_name}")
        pinecone.create_index(name=index_name, metric="cosine", shards=1, dimension=384)

    # Instantiate Pinecone index
    index = pinecone.Index(index_name=index_name)

    return model, index

def add_question_answer_pair(question, plan, df_columns, code, new_rank):
    model, index = init_pinecone()
    # Check if the new rank is above the threshold
    new_rank=int(new_rank)
    if new_rank < 8:
        output_manager.print_wrapper("The new rank is below the threshold. Not adding/updating the vector db record.")
        return

    # Hash the question to be used as id
    id = hashlib.sha256(question.encode()).hexdigest()

    # Vectorize the question
    xq = model.encode([question])[0].tolist()  # Convert the vector to a 1D list

    # Fetch the existing data, if any
    existing_data = index.fetch(ids=[id])
    
    # If the question already exists, get the existing rank
    if existing_data and 'vectors' in existing_data and id in existing_data['vectors']:
        existing_rank = existing_data['vectors'][id]['metadata']['rank']
    else:
        existing_rank = -1

    # If the existing rank is less than the new rank, add or update the question vector and associated data
    if existing_rank < new_rank:
        metadata = {"plan": plan, "df_col": df_columns, "question_txt": question, "code": code, "rank": new_rank}
        vectors = [(id, xq, metadata)]
        index.upsert(vectors=vectors)
        output_manager.print_wrapper(f"Added/Updated the vector db record with id: {id}")
    else:
        output_manager.print_wrapper(f"Existing rank {existing_rank} is higher or equal to the new rank. I am not updating the existing vector db record.")


def retrieve_answer(question, df_columns, match_df=True, similarity_threshold=0.9):
    model, index = init_pinecone()
    # Vectorize the question
    vector = model.encode([question])[0].tolist()  # Convert the vector to a 1D list
    
    # Query the Pinecone index for the closest matching question
    results = index.query(queries=[vector], top_k=1)

    matches = results.get('results', [{}])[0].get('matches', [])

    if not matches:
        output_manager.print_wrapper("No vector db match found")
        return None, None

    match = matches[0]
    closest_match_id = match['id']
    similarity_score = match['score']
    output_manager.print_wrapper(f"Closest match vector db record: {closest_match_id}, Similarity score: {similarity_score}")
    
    # Check if the similarity score is above the threshold
    if similarity_score < similarity_threshold:
        output_manager.print_wrapper(f"Similarity score {similarity_score} is below the threshold {similarity_threshold}")
        return None, None

    fetched_data = index.fetch(ids=[closest_match_id])
    vector_data = fetched_data.get('vectors', {}).get(closest_match_id, {})

    if not vector_data:
        output_manager.print_wrapper("No data found for this vector db id")
        return None, None

    # Get the metadata
    metadata = vector_data['metadata']
    # Check if the dataframe columns match
    if match_df:
        if metadata['df_col'] == df_columns:
            code = metadata['code']
            plan = metadata['plan']
        else:
            output_manager.print_wrapper("The dataframe columns do not match. I will not use this record.")
            return None, None
    # Return the matadata withouth checking the dataframe columns
    else:
        code = metadata['code']
        plan = metadata['plan']

    return code, plan





# prompts.py

# Default Examples (Otherwise Pinecone Long Term Memory)
default_example_output_df = """
Example Output:

```python
import pandas as pd

# Identify the dataframe `df`
# df has already been defined and populated with the required data

# Call the `describe()` method on `df`
df_description = df.describe()

# Print the output of the `describe()` method
print(df_description)
```
"""
default_example_output_gen = """
Example Output:

```python
# Import required libraries
import yfinance as yf
import matplotlib.pyplot as plt

# Define the ticker symbol
tickerSymbol = 'AAPL'

# Get data on this ticker
tickerData = yf.Ticker(tickerSymbol)

# Get the historical prices for this ticker
tickerDf = tickerData.history(period='1d', start='2010-1-1', end='2021-1-1')

# Normalize the data
tickerDf = tickerDf[['Close']]
tickerDf = tickerDf.reset_index()
tickerDf = tickerDf.rename(columns={'Date': 'ds', 'Close': 'y'})

# Plot the close prices
plt.plot(tickerDf.ds, tickerDf.y)
plt.show()
```
"""
default_example_plan_df = """
Example:
```yaml
plan:
  - "Step 1: Convert the 'datetime(GMT)' ..."
  - "Step 2: Calculate the total..."
  - "Step 3: Calculate the pace..."
  - ...
```
"""

default_example_plan_gen = """
Example:
```yaml
plan:
  - "Step 1: Import the yfinance......"
  - "Step 2: Define the ticker..."
  - "Step 3: Download data..."
  - ...
  ```
"""
# Expert Selector Agent Prompts
#   - An 'Internet Research Specialist' that can search the internet to find factual information, relevant data, and contextual details to help address user questions.
expert_selector_system = """
You are a classification expert, and your job is to classify the given task.

1. Determine whether the solution will require an access to a dataset that contains various data, related to the question.
2. Select an expert best suited to solve the task, based on the outcome of the previous step.
   The experts you have access to are as follows:
   - A 'Data Analyst' that can deal with any questions that can be directly solved with code.
   - A 'Data Analysis Theorist' that can answer any general questions on any subject that do not require coding, calculations or data manipulations.    
3. State your level of confidence that if presented with this task, you would be able to solve it accurately and factually correctly on a scale from 0 to 10. Output a single integer.

Formulate your response as a JSON string, with 3 fields {requires_dataset (true or false}, expert, confidence}. Always enclose the JSON string within ```json tags

Example Query:
How many rows are there in this dataset ?

Example Output:
```json
{
  "requires_dataset": true,
  "expert": "Data Analyst",
  "confidence": 10
}
"""
expert_selector_user = """
The user asked the following question: '{}'.
"""
# Analyst Selector Agent Prompts
analyst_selector_system = """
You are a classification expert, and your job is to classify the given task.

1. Select an analyst best suited to solve the task.
  The analysts you have access to are as follows:

    - A 'Data Analyst DF':
      Select this expert if user provided a dataframe. The DataFrame 'df' is already defined and populated with necessary data.

    - A 'Data Analyst Generic':
      Select this expert if user did not provide the dataframe.

2. Rephrase the query, focusing on incorporating previous context and ensuring accuracy in spelling, syntax, and grammar. Make sure that you capture all nuances of the user query.
  If there is a previous context, please put greatest emphasis on the query preceding this one. 
  The rephrased version should be both as descriptive as possible while concise, suitable for later conversion into a detailed, multi-step action plan.
  There is no need to include the dataframe details in the rephrased query.

Formulate your response as a JSON string, with 2 fields {analyst,rephrased_query}. Always enclose the JSON string within ```json tags

Example Query:
How many rows are there in this dataset ?

Example Output:
```json
{
  "analyst": "Data Analyst DF",
  "rephrased_query": "How many rows does this dataset contain?",
}
"""
analyst_selector_user = """
The user asked the following question: '{}', and provided the following dataframe: '{}'.
"""
# Theorist Agent Prompts
theorist_system = """
You are an AI data analysis theorist and your job is to educate the user.
Provide factual information responding directly to the user's question. Include key details and context to ensure your response comprehensively answers their query.

The user asked the following question: '{}'.
"""
# Planner Agent Prompts
planner_system = """
You are an AI assistant capable of assisting users with various tasks related to research, coding, and data analysis. 
The user will inform you about the expertise required to accomplish their task.
You have access to a Google search tool and can retrieve any information that might enhance the analysis.

Today's Date is: {}
"""
planner_user_df = """
Your job is to assist the user with data analysis.

Formulate your response as an algorithm, breaking the solution in up to eight simple, clear English steps, including any values necessary to answer the question.
If fewer steps suffice, that's acceptable. Remember to explain steps rather than write code.

This algorithm will be later converted to Python code and applied to the pandas DataFrame 'df'. 
The result of `print(df.dtypes)` is: 
{}
The DataFrame 'df' is already defined and populated with data!
Explore the content of the dataset to determine whether it is likely to contain the information required to answer the user's question. 
If the dataset is insufficient, you may need to search the web for the missing information.
Formulate your response as a YAML string. Always enclose the YAML string within ```yaml tags.

Allways make sure to incorporate any details or context from the previous conversations, that might be relevant to the task at hand

{}

The user asked the following question: '{}'.
"""
planner_user_gen = """
Your job is to assist the user with data analysis.

Formulate your response as an algorithm, breaking the solution in up to eight simple, clear English steps. You MUST include any values, links or URLs necessary to answer the question!
If fewer steps suffice, that's acceptable. Remember to explain steps rather than write code.

This algorithm will be later converted to Python code.
Formulate your response as a YAML string. Always enclose the YAML string within ```yaml tags.

Allways make sure to incorporate any details or context from the previous conversations, that might be relevant to the task at hand.

{}

The user asked the following question: '{}'.
"""
# Code Generator Agent Prompts
code_generator_system_df = """
You are an AI data analyst and your job is to assist users with analyzing data in the pandas dataframe.
The user will provide a dataframe named `df`, and a list of tasks to be accomplished using Python.
The dataframe df has already been defined and populated with the required data.
"""
code_generator_system_gen = """
You are an AI data analyst and your job is to assist users with data analysis,
or any other tasks related to coding. 
You have not been provided with any datasets, but you have access to the internet.
The user will provide a list of tasks to be accomplished using Python.  
"""
code_generator_user_df = """
You have been presented with a pandas dataframe named `df`.
The dataframe df has already been defined and populated with the required data!
The result of `print(df.dtypes)` is:
{}.
Return the python code that accomplishes the following task: 
{}
Strictly adhere to the prescribed instructions to avoid oversights and ensure an accurate solution.
For context, here is the output of the previous task:
{}
Always include the import statements at the top of the code.
Always include print statements to output the results of your code.
Please make sure that you output contains a FULL, COMPLETE CODE that includes all steps, and solves the task!

{}
"""
code_generator_user_gen = """
Return the python code that accomplishes the following task:
{}
Strictly adhere to the prescribed instructions to avoid oversights and ensure an accurate solution.
For context, here is the output of the previous task:
{}
Always include the import statements at the top of the code.
Always include print statements to output the results of your code.
Please make sure that you output contains a FULL, COMPLETE CODE that includes all steps, and solves the task!

{}
"""
# Error Corrector Agent Prompts
error_corector_system = """
The execution of the code that you provided in the previous step resulted in an error.
Return a complete, corrected python code that incorporates the fixes for the error.
Always include the import statements at the top of the code, and comments and print statements where necessary.

The error message is: {}
"""
# Code Debugger Prompts
code_debugger_system = """
Your job as an AI QA engineer involves correcting and refactoring of the given Code so it delivers the outcome as described in the given Task list.

Code:
{}.
Task list:
{}.

Please follow the below instructions to accomplish your assingment.If provided, the dataframe df has already been defined and populated with the required data.

Task Inspection:
Go through the task list and the given Python code side by side.
Ensure that each task in the list is accurately addressed by a corresponding section of code. 
Do not move on to the next task until the current one is completely solved and its implementation in the code is confirmed.

Code Sectioning and Commenting:
Based on the task list, divide the Python code into sections. Each task from the list should correspond to a distinct section of code.
At the beginning of each section, insert a comment or header that clearly identifies the task that section of code addresses. 
This could look like '# Task 1: Identify the dataframe df' for example.
Ensure that the code within each section correctly and efficiently completes the task described in the comment or header for that section.

After necessary modifications, provide the final, updated code, and a brief summary of the changes you made.
Always use the backticks to enclose the code.

Example Output:
```python
import pandas as pd

# Task 1: Identify the dataframe `df`
# df has already been defined and populated with the required data

# Task 2: Call the `describe()` method on `df`
df_description = df.describe()

# Task 3: Print the output of the `describe()` method
print(df_description)
```
"""
# Code Ranker Agent Prompts
code_ranker_system = """
As an AI QA Engineer, your role is to evaluate and grade the code: {}, supplied by the AI Data Analyst. You should rank it on a scale of 1 to 10.

In your evaluation, consider factors such as the relevancy and accuracy of the obtained results: {} in relation to the original assignment: {},
clarity of the code, and the completeness and format of outputs.

For most cases, your ranks should fall within the range of 5 to 7. Only exceptionally well-crafted codes that deliver exactly as per the desired outcome should score higher. 

Please enclose your ranking in <rank></rank> tags.

Example Output:
<rank>6</rank>
"""
# Solution Summarizer Agent Prompts
solution_summarizer_system = """
The user presented you with the following question.
Question: {}

To address this, you have designed an algorithm.
Algorithm: {}.

You have crafted a Python code based on this algorithm, and the output generated by the code's execution is as follows.
Output: {}.

Please provide a summary of insights achieved through your method's implementation.
Present this information in a manner that is both clear and easy to understand.
Ensure that all results from the computations are included in your summary.
If the user asked for a particular information that is not included in the code execution results, and you know the answer please incorporate the answer to your summary.
"""
# Google Search Query Generator Agent Prompts
google_search_query_generator_system = """
You are an AI internet research specialist and your job is to formulate a user's question as a search query.
Reframe the user's question into a search query as per the below examples.

Example input: Can you please find out what is the popularity of Python programming language in 2023?
Example output: Popularity of Python programming language in 2023

The user asked the following question: '{}'.
"""
# Google Search Summarizer Agent Prompts
google_search_summarizer_system = """
Summarise the below text into an answer for the following question:
  
Question: {}

Present this information in the most clear and comprehensible manner
Be certain to incorporate all relevant facts and insights.
Fill in any information that user has asked for, and that is missing from the text.

Text:

{}
"""
google_search_react_system = """
You are an Internet Research Specialist, and run in a loop of Thought, Action, Observation. This Thought, Action, Observation loop is repeated until you output an Answer.
At the end of the loop you output an Answer.
Use Thought to describe your thoughts about the question you have been asked.
Use Action to run one of the actions available to you.
Observation will be the result of running those actions.

Your available actions are:

calculate:
e.g. calculate: 4 * 7 / 3
Runs a calculation and returns the number - uses Python so be sure to use floating point syntax if necessary

google_search:
e.g. google_search: Popularity of the Python programming language in 2022
Returns a summary of a Google Search
Today's Date is: {}

Use Google Search ONLY if you dont know the answer to the question!

Example session:

Question: What is Leonardo di Caprio's girlfriends age raised to the power of 2?
Thought: I need to search for Leonardo DiCaprio's girlfriend's name.
Action: google_search: Leonardo DiCaprio's girlfriend's name

You will be called again with this:

Observation: Leonardo DiCaprio has had a string of high-profile relationships over the years, including with models Gisele Bündchen, Bar Refaeli, and Nina Agdal. As of 2023, he is currently dating actress and model Camila Morrone.

You then output:

Thought: Camila Morrone's age.
Action: google_search: Camila Morrone's age

You will be called again with this:

Observation: Camila Morrone is 23 years old.

You then output:

Thought: Camila Morrone is 23 years old. I need to raise 23 to the power of 2.
Action: calculate: 23**2

You will be called again with this:

Observation: 529

You then output the finall answer:

Answer: Leonardo's current girlfriend is Camila Morrone, who is 23 years old. 23 raised to the power of 2 is 529.
""".strip()


import os
import time
from anthropic import Client

client = Client()

try:
    # Attempt package-relative import
    from . import output_manager
except ImportError:
    # Fall back to script-style import
    import output_manager

output_manager = output_manager.OutputManager()

def init():
    API_KEY = os.environ.get('ANTHROPIC_API_KEY')
    client.api_key = API_KEY

def convert_openai_to_anthropic(messages):
    updated_data = []
    system_content = ""
    for item in messages:
        if item['role'] == 'system':
            system_content = item['content']
            continue 
        updated_data.append(item)

    return updated_data, system_content

def llm_call(messages: str,model: str,temperature: str,max_tokens: str):  

    init()

    messages, system_instruction = convert_openai_to_anthropic(messages)

    start_time = time.time()

    response = client.messages.create(
        model=model, 
        system=system_instruction,
        messages=messages,
        temperature=temperature,
        max_tokens=max_tokens,
    )

    end_time = time.time()

    elapsed_time = end_time - start_time

    content = response.content[0].text

    prompt_tokens_used = response.usage.input_tokens
    completion_tokens_used = response.usage.output_tokens
    total_tokens_used = prompt_tokens_used + completion_tokens_used

    if elapsed_time > 0:
        tokens_per_second = completion_tokens_used / elapsed_time
    else:
        tokens_per_second = 0

    return content, messages, prompt_tokens_used, completion_tokens_used, total_tokens_used, elapsed_time, tokens_per_second

def llm_stream(log_and_call_manager, chain_id: str,messages: str,model: str,temperature: str,max_tokens: str,tools: str = None):  

    collected_messages = []

    init()

    messages, system_instruction = convert_openai_to_anthropic(messages)

    response = client.messages.create(
        model=model, 
        system=system_instruction,
        messages=messages,
        temperature=temperature,
        max_tokens=max_tokens,
        stream =True,
    )

    start_time = time.time()
    for chunk in response:
        if chunk.type == 'content_block_delta':
            chunk_message = chunk.delta.text
            collected_messages.append(chunk_message)
            output_manager.print_wrapper(chunk_message, end='', flush=True)
        elif chunk.type == 'message_delta':
            completion_tokens_used = chunk.usage.output_tokens
        elif chunk.type == 'message_start':
            prompt_tokens_used = chunk.message.usage.input_tokens


    end_time = time.time()
    elapsed_time = end_time - start_time

    output_manager.print_wrapper("")

    full_reply_content = ''.join([m for m in collected_messages])

    # calculate the total tokens used
    total_tokens_used = prompt_tokens_used + completion_tokens_used
    
    if elapsed_time > 0:
        tokens_per_second = completion_tokens_used / elapsed_time
    else:
        tokens_per_second = 0

    return full_reply_content, messages, prompt_tokens_used, completion_tokens_used, total_tokens_used, elapsed_time, tokens_per_second